mod op_codes;

use op_codes::*;
use std::alloc::{alloc_zeroed, dealloc, Layout};

struct DmObject {
    pointer: *mut u8,
    size: usize,
    layout: Layout,
}

macro_rules! get_32_le {
    ( $code: expr, $i: expr, $offset: literal, $T: ident ) => {
        $code[$i + $offset] as $T
            + (($code[$i + $offset + 1] as $T) << 8)
            + (($code[$i + $offset + 2] as $T) << 16)
            + (($code[$i + $offset + 3] as $T) << 24)
    };
}

macro_rules! get_64_le {
    ( $code: expr, $i: expr, $offset: literal, $T: ident ) => {
        $code[$i + $offset] as $T
            + (($code[$i + $offset + 1] as $T) << 8)
            + (($code[$i + $offset + 2] as $T) << 16)
            + (($code[$i + $offset + 3] as $T) << 24)
            + (($code[$i + $offset + 4] as $T) << 32)
            + (($code[$i + $offset + 5] as $T) << 40)
            + (($code[$i + $offset + 6] as $T) << 48)
            + (($code[$i + $offset + 7] as $T) << 56)
    };
}

pub fn run(code: &Vec<u8>, registers: &mut Vec<u64>, register_types: &mut Vec<RegisterType>) {
    let mut i = 0;
    while i < code.len() {
        match code[i] {
            MOV_INT => {
                let target_register = code[i + 1] as usize;
                let operand = get_32_le!(code, i, 2, u32);
                registers[target_register] = operand as u64;
                register_types[target_register] = RegisterType::Int;
                i += 6;
            }
            MOV_LONG => {
                let target_register = code[i + 1] as usize;
                let operand = get_64_le!(code, i, 2, u64);
                registers[target_register] = operand;
                register_types[target_register] = RegisterType::Long;
                i += 10;
            }
            MOV_DOUBLE => { /* todo */ }
            MOV_REGISTER => {
                let target_register = code[i + 1] as usize;
                let source_register = code[i + 2] as usize;
                registers[target_register] = registers[source_register];
                register_types[target_register] = register_types[source_register].clone();
                i += 3;
            }
            ALLOC => {
                let target_register = code[i + 1] as usize;
                let size = get_32_le!(code, i, 2, usize);
                let layout = Layout::from_size_align(size, 4).unwrap();
                let pointer = unsafe { alloc_zeroed(layout) };
                let dm_object = Box::new(DmObject {
                    pointer,
                    size,
                    layout,
                });
                let dm_object_pointer = Box::into_raw(dm_object) as u64;
                registers[target_register] = dm_object_pointer;
                register_types[target_register] = RegisterType::Pointer;
                i += 6;
            }
            DEALLOC => {
                let target_register = code[i + 1] as usize;
                let box_address = registers[target_register];
                unsafe {
                    let dm_object = Box::from_raw(box_address as *mut DmObject);
                    dealloc(dm_object.pointer, dm_object.layout);
                }
                registers[target_register] = 0;
                register_types[target_register] = RegisterType::Int;
                i += 2;
            }
            MOV_INT_TO => {
                let target_register = code[i + 1] as usize;
                if register_types[target_register] != RegisterType::Pointer {
                    panic!("target_register {} is not a Pointer", target_register);
                }
                let offset = get_32_le!(code, i, 2, isize);
                let box_address = registers[target_register];
                let new_address = unsafe {
                    let dm_object = Box::from_raw(box_address as *mut DmObject);
                    let pointer = dm_object.pointer;
                    pointer.offset(offset).write(code[i + 6]);
                    pointer.offset(offset + 1).write(code[i + 7]);
                    pointer.offset(offset + 2).write(code[i + 8]);
                    pointer.offset(offset + 3).write(code[i + 9]);
                    Box::into_raw(dm_object) as u64
                };
                registers[target_register] = new_address;
                i += 10;
            }
            _ => panic!("Invalid code instruction"),
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use std::alloc::dealloc;

    fn init_registers(n_registers: usize) -> (Vec<u64>, Vec<RegisterType>) {
        (vec![0; n_registers], vec![RegisterType::Int; n_registers])
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(1, registers[0]);
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(0xffff, registers[0]);
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff_ffff);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(0xffff_ffff, registers[0]);
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_register(&mut code, 0, 1);
        let (mut registers, mut register_types) = init_registers(2);
        registers[1] = 1;
        run(&code, &mut registers, &mut register_types);
        assert_eq!(registers[0], 1);
    }

    #[test]
    fn mov_int_to_register_as_address() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_mov_int_to(&mut code, 0, 0, 0xff);
        let mut registers = vec![0; 16];
        let mut register_types = vec![RegisterType::Int; 16];
        run(&code, &mut registers, &mut register_types);
        let box_address = registers[0];
        unsafe {
            let dm_object = Box::from_raw(box_address as *mut DmObject);
            assert_eq!(0xff, *dm_object.pointer);
            dealloc(dm_object.pointer, dm_object.layout);
        }
    }

    #[test]
    fn alloc_and_dealloc_expect_register_cleared() {
        let mut code = Vec::new();
        add_alloc(&mut code, 0, 4);
        add_dealloc(&mut code, 0);
        let (mut registers, mut register_types) = init_registers(1);
        run(&code, &mut registers, &mut register_types);
        assert_eq!(0, registers[0]);
        assert_eq!(RegisterType::Int, register_types[0]);
    }
}
